#!/usr/bin/env python3
# Version: 1.0.19, build date: 2021-06-16 11:04:35.293294
import json
import requests
import urllib3
from urllib.parse import parse_qs, urlencode, urlsplit, urlunsplit

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

class Kameleoon:
    client_id = ''
    client_secret = ''
    api_url = 'https://api.kameleoon.com'
    site_codes = []
    token_url = api_url + '/oauth/token'
    kameleoon_configuration_path = '/etc/kameleoon/'
    kameleoon_configuration_file = 'configuration.json'
    headers = {}

    def __init__(self):
        try:
            with open(self.getKameleoonConfigurationPathFile()) as json_file:
                kameleoon_auth = json.load(json_file)
            if 'api_url' in kameleoon_auth:
                self.api_url = kameleoon_auth['api_url']
            self.site_codes = kameleoon_auth['site_codes']
            self.client_id = kameleoon_auth['client_id']
            self.client_secret = kameleoon_auth['client_secret']
        except:
            print("Error while getting configuration from: " + self.getKameleoonConfigurationPathFile())

    def getKameleoonConfigurationPathFile(self):
        return self.kameleoon_configuration_path+self.kameleoon_configuration_file

    def getHeaders(self):
        if not self.headers:
            self.headers = {
                'Authorization': 'Bearer ' + self.getToken(),
                'Content-Type': 'application/json'
            }
        return self.headers

    def getToken(self):
        try:
            headers = {'Content-Type': 'application/x-www-form-urlencoded'}
            payload = 'grant_type=client_credentials' \
                      + '&client_id=' + self.client_id \
                      + '&client_secret=' + self.client_secret
            response = requests.request("POST", self.token_url, headers=headers, data=payload)
            return response.json()['access_token']
        except:
            print('Error with api authorization')
            raise

    def getSiteIds(self):
        try:
            sites_id = []
            sites = self.fetchAllResults(self.api_url + '/sites')
            for site_code_to_find in self.site_codes:
                site_id_found = None
                for site_from_response in sites:
                    if site_code_to_find == site_from_response['code']:
                        site_id_found = site_from_response['id']
                        continue
                if site_id_found is None:
                    print("Unable to retrieve the id for site code: " + site_code_to_find)
                    raise
                sites_id.append(site_id_found)
            return sites_id
        except:
            print('Error with api to sites')
            raise

    def getParams(self, current_page, filters):
        params = {"page": current_page, "perPage": -1}
        if filters:
            params.update({"filter":json.dumps(filters, separators=(',', ':'))})
        return params

    def getCurrentPage(self, response):
        return int(response.headers["X-Pagination-Current-Page"])

    def getPageCount(self, response):
        return int(response.headers["X-Pagination-Page-Count"])

    def fetchOneResult(self, url):
        try:
            response = requests.get(
                url,
                headers= self.getHeaders(),
                verify=False
            )
            response.raise_for_status()
            return response.json()
        except requests.exceptions.HTTPError as e:
            print('Error while fetching results:', e)
            raise
        except:
            print('Error while fetching results')
            raise

    def fetchAllResults(self, url, filters=None):
        try:
            results = []
            current_page = 1
            while "There are still results":
                response = \
                    requests.get(
                        url,
                        params=self.getParams(current_page, filters),
                        headers=self.getHeaders(),
                        verify=False
                    )
                response.raise_for_status()
                results.extend(response.json())
                if self.getPageCount(response) == 0 or self.getCurrentPage(response) == self.getPageCount(response):
                    break
                else:
                    current_page += 1
            return results
        except requests.exceptions.HTTPError as e:
            print('Error while fetching results:', e)
            raise
        except:
            print('Error while fetching results')
            raise

    def getExperiments(self, sites_id):
        try:
            experiments = self.fetchAllResults(
                self.api_url + '/experiments',
                [
                    {'field':'type','operator':'IN','parameters':['SERVER_SIDE','HYBRID']},
                    {'field':'status','operator':'IN','parameters':['ACTIVE', 'DEVIATED']},
                    {'field':'siteId','operator':'IN','parameters':sites_id}
                ]
            )
            return filter(lambda experiment: 'targetingConfiguration' in experiment, experiments)
        except:
            print('Error with api to experiments')
            raise

    def getSegment(self, segment_id):
        return self.fetchOneResult(self.api_url + '/segments/' + str(segment_id))

    def getVariationsConfiguration(self, experiment):
        variationsConfig = []
        for variationId in experiment['variationsId']:
            variation = self.getVariation(variationId)
            variationLite = {}
            if 'redirection' in variation.keys():
                variationLite['id'] = variation['id']
                variationLite['type'] = variation['redirection']['type']
                variationLite['redirection'] = self.get_redirection(variation['redirection'], experiment['baseURL'])
                if (str(variation['id']) in experiment['deviations']) and (variation['redirection'] != ""):
                    variationLite['weight'] = experiment['deviations'][str(variation['id'])]
                    variationsConfig.append(variationLite)
        return variationsConfig

    def getVariation(self, variation_id):
        #TODO optimize with one request to avoid N+1 requests.
        return self.fetchOneResult(self.api_url + '/variations/' + str(variation_id))

    @staticmethod
    def add_query_parameters(query, new_query):
        query_dict = parse_qs(query)
        query_dict.update(parse_qs(new_query))
        return urlencode(query_dict, doseq=True)

    def get_query_parameters_global_redirection(self, redirection_params, base_url):
        path = urlsplit(redirection_params['url']).path
        query = urlsplit(redirection_params['url']).query
        fragment = urlsplit(redirection_params['url']).fragment
        if redirection_params['includeQueryParameters'] is True:
            query = self.add_query_parameters(query, urlsplit(base_url).query)
        return urlunsplit(('', '', path, query, fragment))

    def get_redirection(self, redirection_params, base_url):
        try:
            if redirection_params['type'] == 'GLOBAL_REDIRECTION':
                return self.get_query_parameters_global_redirection(redirection_params, base_url)
            elif redirection_params['type'] == 'PARAMETER_REDIRECTION':
                return redirection_params['parameters']
            else:
                print('Unknown redirection type: ', redirection_params['type'])
                raise
        except:
            print('Error to get redirection')
            raise

    def getValueOfTargetingConfiguration(self, experiment):
        mapTargetingConfigurationToValue = {
            'SITE': 'S',
            'PAGE': 'P',
            'URL': 'U',
            'SAVED_TEMPLATE': 'T'
        }
        return mapTargetingConfigurationToValue[experiment['targetingConfiguration']]

    @staticmethod
    def byPageUrl(condition):
        return True if condition['targetingType'] == 'PAGE_URL' else False

    @staticmethod
    def formatInclude(condition):
        condition['include'] = ('Y' if condition['include'] else 'N')
        return condition

    def getConditionsData(self, segmentId):
        response = []
        conditionsData = self.getSegment(segmentId)['conditionsData']
        i = -1
        for level in conditionsData['firstLevel']:
            level['orOperator'] = ('' if i == -1 else ('|' if conditionsData['firstLevelOrOperators'][i] else '&'))
            i += 1
            j = 1
            for orOperator in level['orOperators']:
                level['conditions'][j]['orOperator'] = ('|' if orOperator else '&')
                j += 1
            level['conditions'] = list(filter(self.byPageUrl, level['conditions']))
            level['conditions'] = list(map(self.formatInclude, level['conditions']))
            level.pop('orOperators')
            if len(level['conditions']):
                level['conditions'][0]['orOperator'] = ''
                response.append(level)
        return response

    def getJsonReady(self):
        kameleoon_json = []
        try:
            for experiment in self.getExperiments(self.getSiteIds()):
                experimentConfig = {}
                experimentConfig['targetingConfiguration'] = self.getValueOfTargetingConfiguration(experiment)
                if experimentConfig['targetingConfiguration'] != 'T':
                    break
                experimentConfig['baseUrl'] = experiment['baseURL']
                experimentConfig['experimentId'] = experiment['id']
                experimentConfig['conditionsData'] = self.getConditionsData(experiment['targetingSegmentId'])
                experimentConfig['variations'] = self.getVariationsConfiguration(experiment)
                kameleoon_json.append(experimentConfig)
            return kameleoon_json
        except:
            print("Error while getting config in JSON.")
            raise


kameleoon_json = '/tmp/kameleoon-data.json'
default_nginx_conf_file = '/etc/nginx/modules/00-kameleoon.conf'

# Get configuration from api in json format
jsonConf = []
kameleoonConf= []
try:
    k = Kameleoon()
    with open(k.getKameleoonConfigurationPathFile()) as json_file:
        kameleoonConf = json.load(json_file)
    if 'nginxConfigurationFile' not in kameleoonConf:
        reply = str(input("Do you agree with the default path for the module configuration file: "+default_nginx_conf_file+"? (y/n): ")).lower().strip()
        if reply[0] == 'y':
            kameleoonConf['nginxConfigurationFile'] = default_nginx_conf_file
        else:
            kameleoonConf['nginxConfigurationFile'] = str(input("Type your nginx configuration file: ")).strip()
        with open(k.getKameleoonConfigurationPathFile(), 'w') as outfile:
            json.dump(kameleoonConf, outfile)
    jsonConf = k.getJsonReady()

except ValueError as error:
    print("Error while getting configuration:", error.strerror)
    exit()

# Generates mod_kameleoon configuration file
try:
    with open(kameleoon_json, 'w') as f:
        json.dump(jsonConf, f)
except:
    print('Failed to generate '+kameleoon_json+' file. Do you have access right for /tmp/ ?')
    exit()

# Generates nginx configuration file
try:
    with open(kameleoonConf['nginxConfigurationFile'], 'w') as f:
        for experiment in jsonConf:
            for variation in experiment['variations']:
                f.write('''
location /%s {
    ''' % ("experiment_"+str(experiment['experimentId'])+"_variation_"+str(variation['id'])))
                if variation['type'] == 'GLOBAL_REDIRECTION':
                    f.write('''
    rewrite ^(/.*)$ %s last;
    ''' % (variation['redirection']))
                else:
                    f.write('''
    if ($is_args = "?") {
        rewrite ^ ${request_uri};
        rewrite ^([^?]*)(\?.*)$ $1?${args}&%s? last;
    }
    if ($is_args != "?") {
        rewrite ^(/.*) ${request_uri}?%s last;
    }
    ''' % (variation['redirection'], variation['redirection']))
                f.write('''
}
    ''')
            print('Finished with success.')
except:
    print('Failed to generate nginx configuration file. Does the following path exist: '+kameleoonConf['nginxConfigurationFile']+' ?')
